#!/usr/bin/perl

if ($#ARGV != 1)
{
	die "usage: formatbnf.pl <language> <input-file>\n" .
		"Where <language> is one of bnf, mif, lout\n";
}

$language = $ARGV[0];
$infile = $ARGV[1];

open (INFILE, "<$infile") or die "Can't open file `$infile'\n";

%printFunctions = (
	"mif" => [MIFPrologue, MIFRule, MIFFirstAlternative, MIFAlternative, Nop, MIFEpilogue,
		MIFSpace, MIFNonterminal, MIFTerminal, MIFOpen, MIFClose, MIF0OrMoreClose,
		MIF0Or1Close, MIF1OrMoreClose, MIFOr, MIFEllipsis],
	"bnf" => [Nop, BNFRule, BNFFirstAlternative, BNFAlternative, Nop, Nop,
		BNFSpace, BNFNonterminal, BNFTerminal, BNFOpen, BNFClose, BNF0OrMoreClose,
		BNF0Or1Close, BNF1OrMoreClose, BNFOr, BNFEllipsis],
	"lout" => [Nop, LoutRule, LoutFirstAlternative, LoutAlternative, NL, Nop,
		BNFSpace, LoutNonterminal, LoutTerminal, BNFOpen, BNFClose, BNF0OrMoreClose,
		BNF0Or1Close, BNF1OrMoreClose, LoutOr, BNFEllipsis],
);

$functions = $printFunctions{$language};

if (! defined $printFunctions{$language} )
{
	die "formatbnf.pl doesn't support language `$language'\n";
}

sub Nop { }
sub NL { print "\n"; }

sub PrintPrologue { $functions->[0] (@_); }
sub PrintRule { $functions->[1] (@_); }
sub PrintFirstAlternative { $functions->[2] (@_); }
sub PrintAlternative { $functions->[3] (@_); }
sub PrintCloseRule { $functions->[4] (@_); }
sub PrintEpilogue { $functions->[5] (@_); }

sub PrintTokenSpace { $functions->[6] (@_); }
sub PrintNonterminal { $functions->[7] (@_); }
sub PrintTerminal { $functions->[8] (@_); }
sub PrintOpen { $functions->[9] (@_); }
sub PrintClose { $functions->[10] (@_); }
sub Print0OrMoreClose { $functions->[11] (@_); }
sub Print0Or1Close { $functions->[12] (@_); }
sub Print1OrMoreClose { $functions->[13] (@_); }
sub PrintOr { $functions->[14] (@_); }
sub PrintEllipsis { $functions->[15] (@_); }

# Array of [nameOfNonTerminal alternate1 alternate2 ...]
@nonTerms = ();

$nonTermName = "";
@alternatives = ();
$thisAlternative = "";

while (<INFILE>)
{
	if (/^<([\w-]+)>\s*::=\s*(.*)/)
	{
		if ($nonTermName ne "") { EndOfRule (); }
		$nonTermName = $1;
		$endOfLine = $2;
		$nonTermName =~ s/-/_/g;
		$endOfLine =~ s/\s+//g;
		$thisAlternative = $endOfLine;
	} elsif (/^\|\s*(.*)/)
	{
		$endOfLine = $1;
		$endOfLine =~ s/\s+//g;
		FinishAlternative ();
		$thisAlternative = $endOfLine;
	} elsif (/^[ \t]/)
	{ # Line continuation
		$endOfLine = $_;
		$endOfLine =~ s/\s+//g;
		$thisAlternative .= "\n" . $endOfLine;
	} else
	{ # End of non terminal or irrelevant line
		EndOfRule ();
	}
}

sub EndOfRule
{
	if ($nonTermName ne "")
	{
		FinishAlternative ();
		push (@nonTerms, [$nonTermName, @alternatives]);

		$nonTermName = "";
		@alternatives = ();
	}
}

sub FinishAlternative # Handle last line of alternative
{
	if ($thisAlternative ne "")
	{		
		push (@alternatives, $thisAlternative);
		$thisAlternative = "";
	}
}

PrintPrologue ();

for $i (@nonTerms)
{
	@nonTerm = @{$i};

	PrintRule ($nonTerm[0]);
	PrintFirstAlternative ($nonTerm[1]);
	if ($#nonTerm > 1)
	{
		for $altNo (2..$#nonTerm)
		{
			PrintAlternative ($nonTerm[$altNo])
		}
	}
	PrintCloseRule ();
}

sub FormatRuleLine
{
	my $line = shift;

	do {
		$foundToken = 0;
		if ($line =~ s/^re(.)([^\1]*?)\1(.*)/$3/)
		{
			$token = $2;
			PrintRegexp ($token);
			$foundToken = 1;

		} elsif ($line =~ s/^<([\w-]+)>(.*)/$2/)
		{
			$token = $1;
			$token =~ s/-/_/g;
			PrintNonterminal ($token);
			$foundToken = 1;
		} elsif ($line =~ s/^"([^"]*)"(.*)/$2/ or
			$line =~ s/^`([^']*)'(.*)/$2/)
		{
			$token = $1;
			# Put back a space if you find `' or "" as this
			# must have been ` ' or " "
			# AB
			$token =~ s/^$/ /;

			PrintTerminal ($token);
			$foundToken = 1;
		} elsif ($line =~ s/^\((.*)/$1/)
		{
			$token = $1;
			PrintOpen ();
			$foundToken = 1;
		} elsif ($line =~ s/^\)\*(.*)/$1/)
		{
			$token = $1;
			Print0OrMoreClose ();
			$foundToken = 1;
		} elsif ($line =~ s/^\)\+(.*)/$1/)
		{
			$token = $1;
			Print1OrMoreClose ();
			$foundToken = 1;
		} elsif ($line =~ s/^\)\?(.*)/$1/)
		{
			$token = $1;
			Print0Or1Close ();
			$foundToken = 1;
		} elsif ($line =~ s/^\)(.*)/$1/)
		{
			$token = $1;
			PrintClose ();
			$foundToken = 1;
		} elsif ($line =~ s/^\|(.*)/$1/)
		{
			$token = $1;
			PrintOr ();
			$foundToken = 1;
		} elsif ($line =~ s/^\.\.\.(.*)/$1/)
		{
			$token = $1;
			PrintEllipsis ();
			$foundToken = 1;
		}
		PrintTokenSpace () if ($foundToken);
	} while ($foundToken);
}

PrintEpilogue ();

############## BNF
sub BNFRule
{
	my $arg = shift;

	print "<$arg> ::= ";
}
sub BNFFirstAlternative
{
	my $arg = shift;
	my @lines = split /\n/, $arg;

	FormatRuleLine ($lines[0]);
	print "\n";
	if ($#lines > 0)
	{
		for $lineNo (1..$#lines)
		{
			my $line = $lines[$lineNo];

			print "        ";
			FormatRuleLine ($line);
			print "\n";
		}
	}
}
sub BNFAlternative
{
	my $arg = shift;
	my @lines = split /\n/, $arg;

	print "|   ";
	FormatRuleLine ($lines[0]);
	print "\n";
	if ($#lines > 0)
	{
		for $lineNo (1..$#lines)
		{
			my $line = $lines[$lineNo];

			print "        ";
			FormatRuleLine ($line);
			print "\n";
		}
	}
}
sub BNFSpace { print " "; }
sub BNFNonterminal
{
	my $arg = shift;

	print "<$arg>";
}
sub BNFTerminal
{
	my $arg = shift;

	print "`$arg'";
}
sub BNFOpen { print "("; }
sub BNFClose { print ")"; }
sub BNF0OrMoreClose { print ")*"; }
sub BNF0Or1Close { print ")?"; }
sub BNF1OrMoreClose { print ")+"; }
sub BNFOr { print "|"; }
sub BNFEllipsis { print "..."; }
############## MIF
sub MIFPrologue
{
	$MIFStartCell = "        <Cell
          <CellContent
";
	$MIFEndCell = "          > # end CellContent
        > # end of Cell
";
	$MIFStartPara = "            <Para
            <PgfTag BNFBody>
              <ParaLine
";
	$MIFEndPara = "              > # end of ParaLine
            > # end of Para
";
	$MIFStartString = "              <String `" ;
	$MIFEndString = "'> # end of string
";
# Top/Bottom Cell margins are absolute: Left, Top, Right, Bottom
	$MIFCellMargins = "        <Pgf
          <PgfCellMargins 0 pt 3 pt 0 pt 3 pt>
          <PgfCellTMarginFixed Yes>
          <PgfCellBMarginFixed Yes>
        > # End Pgf
";
	$MIFTerminalFont = "              <Font
                <FFamily `Courier'>
                <FEncoding `FrameRoman'>
                <FWeight `Bold'>
                <FSize 10>
               > # End of Font Defn
";
	$MIFTerminalFont2 = "              <Font
                <FFamily `Courier New'>
                <FEncoding `FrameRoman'>
                <FWeight `Bold'>
                <FSize 10>
               > # End of Font Defn
";
	$MIFNonTerminalFont = "              <Font
                <FPlatformName `M.Helvetica.P'>
                <FFamily `Helvetica'>
                <FWeight `Medium'>
                <FPostScriptName `Helvetica'>
                <FEncoding `FrameRoman'>
                <FSize 9>
               > # End of Font Defn
";

	$MIFSymbolFont = "              <Font
                <FPlatformName `M.Symbol.P'>
                <FFamily `Symbol'>
                <FWeight `Medium'>
                <FPostScriptName `Symbol'>
                <FEncoding `FrameRoman'>
                <FSize 10>
              > # End of Font Defn
";
	$MIFEndFont = "              <Font
                <FTag `'>
              > # End Font Defn
";
	print "<MIFFile 5.50>
<Tbls
  <Tbl
    <TblID 1>
    <TblTag BNF>
    <TblNumColumns 3>
    <TblColumnWidth 3.1cm>
    <TblColumnWidth 0.9cm>
    <TblColumnWidth 10.1cm>
    <TblFormat
      #                L    T    R    B
      <TblCellMargins  6 pt 9 pt 0 pt 3 pt>
    > # End of TblFormat
    <TblBody
";
}
sub MIFEpilogue
{
  print "    > # end of TblBody
  > # end of Tbl
> # end of Tbls
<Para
  <ParaLine
    <String `Insert the table'>
    <ATbl 1> # instantiate the table
  > # End ParaLine
> # End Para\n";
}
sub MIFClearCells
{
	$Col1 = "";
	$Col2 = "";
	$Col3 = "";
}
sub MIFPrintCells
{
  	print "      <Row\n";
	print $Col1, $Col2, $Col3;
    print "      > # End of Row\n";
}
sub MIFRule
{
	my $arg = shift;

	MIFClearCells ();
	$Col1 .= MIFNonterminalString ($arg);
	$Col1 = $MIFStartCell . $MIFStartPara . $Col1 . $MIFEndPara. $MIFEndCell;
}
sub MIFFirstAlternative
{
	my $arg = shift;
	my @lines = split /\n/, $arg;

	$Col2 .= $MIFStartPara . $MIFStartString;
	$Col2 .= "::=";
	$Col2 .= $MIFEndString . $MIFEndPara;

	$Col3 .= $MIFStartPara ;
	FormatRuleLine ($lines[0]);
	$Col3 .= $MIFEndPara;

	if ($#lines > 0)
	{
		for $lineNo (1..$#lines)
		{
		  $Col2 .= $MIFStartPara;
		  $Col2 .= $MIFEndPara;
		  my $line = $lines[$lineNo];
		  $Col3 .= $MIFStartPara . $MIFStartString . "\\t" . $MIFEndString;

		  FormatRuleLine ($line);
		  $Col3 .= $MIFEndPara;
		}
	}
	$Col2  = $MIFStartCell . $Col2 . $MIFEndCell ;
	$Col3  = $MIFStartCell . $Col3 . $MIFEndCell ;
	MIFPrintCells ();
}
sub MIFAlternative
{
	my $arg = shift;
	my @lines = split /\n/, $arg;

	MIFClearCells ();
	$Col2 .= $MIFStartPara .$MIFCellMargins . $MIFStartString;
	$Col2 .= "|" ;
	$Col2 .=  $MIFEndString . $MIFEndPara ;

	$Col3 .= $MIFStartPara . $MIFCellMargins;
	FormatRuleLine ($lines[0]);
	$Col3 .= $MIFEndPara;

	if ($#lines > 0)
	{
		for $lineNo (1..$#lines)
		{
		  $Col2 .= $MIFStartPara . $MIFStartString;
		  $Col2 .= " ";
		  $Col2 .= $MIFEndString . $MIFEndPara;
		  my $line = $lines[$lineNo];
		  $Col3 .= $MIFStartPara . $MIFStartString . "\\t" . $MIFEndString;
		  FormatRuleLine ($line);
		  $Col3 .= $MIFEndPara;
		}
	}
	$Col1 = $MIFStartCell . $MIFStartPara .$MIFCellMargins . $MIFEndPara. $MIFEndCell;
	$Col2  = $MIFStartCell . $Col2 . $MIFEndCell ;
	$Col3  = $MIFStartCell . $Col3 . $MIFEndCell ;
	MIFPrintCells ();
}
sub MIFSpace
{ $Col3 .= $MIFStartString . " " . $MIFEndString; }
sub MIFNonterminalString
{
	my $arg = shift;
	my $str;
	$arg =~ s/_/-/g;

	$str = $MIFSymbolFont . $MIFStartString . "\\xe1 " . $MIFEndString . $MIFEndFont;
	$str .= $MIFNonTerminalFont . $MIFStartString . $arg . $MIFEndString  .$MIFEndFont;
	$str .= $MIFSymbolFont . $MIFStartString . "\\xf1 " .$MIFEndString . $MIFEndFont;

	return $str;
}
sub MIFNonterminal
{
	my $arg = shift;

	$Col3 .= MIFNonterminalString ($arg);
}
sub MIFTerminal
{
	my $arg = shift;
	my $font = $MIFTerminalFont;

	$arg =~ s/'/\\'/g;
	$arg =~ s/([<>])/\\$1/g;

	if ($arg =~ /\|/)
	{
	    $font = $MIFTerminalFont2;
	}

	$Col3 .= $font .
	  $MIFStartString . $arg . $MIFEndString .
	  $MIFEndFont;
}
sub MIFOpen
{ $Col3 .= $MIFStartString . "(" . $MIFEndString; }
sub MIFClose
{ $Col3 .= $MIFStartString .  ")" . $MIFEndString; }
sub MIF0OrMoreClose
{ $Col3 .= $MIFStartString . ")*" . $MIFEndString ; }
sub MIF0Or1Close
{ $Col3 .= $MIFStartString . ")?" .$MIFEndString ; }
sub MIF1OrMoreClose
{ $Col3 .= $MIFStartString . ")+" . $MIFEndString ; }
sub MIFOr
{ $Col3 .=  $MIFStartString . "|" . $MIFEndString ; }
sub MIFEllipsis
{ $Col3  .=  $MIFStartString . "..." . $MIFEndString  ; }
############## Lout
sub LoutRule
{
	my $arg = shift;

	print "\@Rowa marginabove { 0.3c } A {";
	PrintNonterminal ($arg);
	print "} B { ::= } C {";
}
sub LoutFirstAlternative
{
	my $arg = shift;
	my @lines = split /\n/, $arg;

	FormatRuleLine ($lines[0]);
	print "\n";
	if ($#lines > 0)
	{
		for $lineNo (1..$#lines)
		{
			my $line = $lines[$lineNo];

			print "}\n";
			print "\@Rowb A {} B {} C {|2m} D {";
			FormatRuleLine ($line);
		}
	}
	print "}\n";
}
sub LoutAlternative
{
	my $arg = shift;
	my @lines = split /\n/, $arg;

	print "\@Rowa A {} B {\"|\"} C {";
	FormatRuleLine ($lines[0]);
	if ($#lines > 0)
	{
		for $lineNo (1..$#lines)
		{
			my $line = $lines[$lineNo];

			print "}\n";
			print "\@Rowb A {} B {} C {} D {";
			FormatRuleLine ($line);
		}
	}
	print "}\n";
}
sub LoutNonterminal
{
	my $arg = shift;

	$arg =~ s/_/-/g;
	print "\@NonTerm{",$arg,"}";
}
sub LoutTerminal
{
	my $arg = shift;

	if ($arg eq " ")
	{
		print "\@Keyword{\@Eq{square}}";
	} else {
		$arg =~ s/(['"\\])/\\$1/g;
		print "\@Keyword{\"", $arg, "\"}";
	}
}
sub LoutOr { print "\"|\""; }
